#include "benchmark.h"
#include <QtGui/QPainter>

const int width = 500;
const int height = 500;
const QColor main_color = Qt::green;

void Benchmark::hsvGradient()
{
    QImage image = QImage(QSize(width, height), QImage::Format_ARGB32);

    QBENCHMARK
    {
        float h = main_color.hsvHueF();
        for (int s=0; s<width; ++s)
        {
            for (int v=0; v<height; ++v)
            {
                QColor color = QColor::fromHsvF(
                        h,
                        1.0 * s / (width-1),
                        1.0 - (1.0 * v / (height-1)));
                image.setPixel(s, v, color.rgb());
            }
        }
    }

    image.save("hsvGradient.png");
}

void Benchmark::doubleLinearGradient()
{
    QImage image;
    QPixmap pixmap(QSize(width, height));
    QPainter painter(&pixmap);
    painter.setPen(QPen(Qt::NoPen));

    QBENCHMARK
    {
        QLinearGradient h_gradient(QPointF(1.0, 0.0), QPointF(width, 0.0));
        h_gradient.setColorAt(0, Qt::white);
        h_gradient.setColorAt(1, main_color);
        painter.setBrush(QBrush(h_gradient));
        painter.drawRect(QRect(0, 0, width, height));

        QLinearGradient v_gradient(QPointF(1.0, 0.0), QPointF(0.0, height));
        v_gradient.setColorAt(0, Qt::transparent);
        v_gradient.setColorAt(1, Qt::black);
        painter.setBrush(QBrush(v_gradient));
        painter.drawRect(QRect(0, 0, width, height));
    }

    image = pixmap.toImage();
    image.save("doubleLinearGradient.png");
}

void Benchmark::drawInBuffer()
{
    uchar buffer[width * height * 4];
    QImage image(buffer, width, height, QImage::Format_ARGB32);

    QBENCHMARK
    {
        uchar* pbuffer = &buffer[0];

        for (int i=0; i<width; ++i)
        {
            //uchar green_blue = red;
            for (int j=0; j<height; ++j)
            {
                float a = 255.0 * (1.0 - (1.0 * i / width));
                float b = a * (1.0 - (1.0 * j / height));

                *pbuffer = (uchar) b; // blue
                ++pbuffer;

                *pbuffer = (uchar) b; // green
                ++pbuffer;

                *pbuffer = (uchar) a; // red
                ++pbuffer;

                *pbuffer = (uchar) 255; // alpha
                ++pbuffer;
            }
        }
    }

    image.save("buffured.png");
}

void bresenhamVertical(QImage& image, const int x, int red, int green, int blue)
{
    int y = 0;
    const int y_end = height - 1;

    const int red_end = 255;
    const int green_end = 255;
    const int blue_end = 255;

    const int e_y = y_end - y;
    const int d_y = 2 * e_y;

    int e_red = e_y;
    int e_green = e_y;
    int e_blue = e_y;

    const int d_red = 2 * (red_end - red);
    const int d_green = 2 * (green_end - green);
    const int d_blue = 2 * (blue_end - blue);

    while (y <= y_end)
    {
        image.setPixel(QPoint(x, y), qRgb(255 - red, 255 - green, 255 - blue));

        ++y;

        e_red -= d_red;
        if (e_red <= 0)
        {
            ++red;
            e_red += d_y;
        }

        e_green -= d_green;
        if (e_green <= 0)
        {
            ++green;
            e_green += d_y;
        }

        e_blue -= d_blue;
        if (e_blue <= 0)
        {
            ++blue;
            e_blue += d_y;
        }
    }
}

void bresenhamHorizontal(QImage& image)
{
    int x = 0;
    const int x_end = width - 1;

    int red = 0;
    int green = 0;
    int blue = 0;

    const int red_end = 255 - main_color.red();
    const int green_end = 255 - main_color.green();
    const int blue_end = 255 - main_color.blue();

    const int e_x = x_end - x;
    const int d_x = 2 * e_x;

    int e_red = e_x;
    int e_green = e_x;
    int e_blue = e_x;

    const int d_red = 2 * (red_end - red);
    const int d_green = 2 * (green_end - green);
    const int d_blue = 2 * (blue_end - blue);

    while (x <= x_end)
    {
        bresenhamVertical(image, x, red, green, blue);

        ++x;

        e_red -= d_red;
        if (e_red <= 0)
        {
            ++red;
            e_red += d_x;
        }

        e_green -= d_green;
        if (e_green <= 0)
        {
            ++green;
            e_green += d_x;
        }

        e_blue -= d_blue;
        if (e_blue <= 0)
        {
            ++blue;
            e_blue += d_x;
        }
    }
}

void Benchmark::bresenham()
{
    // version monochrome
    {
        /*
        QImage image(width, height, QImage::Format_ARGB32);

        int x1 = 0; // premier pixel
        int y1 = 0; // composante dans le premier pixel
        const int x2 = width - 1; // dernier pixel
        const int y2 = 255; // composante dans le dernier pixel

        int e = x2 - x1;
        const int dx = 2 * e;
        const int dy = 2 * (y2 - y1);

        while (x1 <= x2)
        {
            for (int i=0; i<height; ++i)
                image.setPixel(QPoint(x1, i), qRgb(y1, y1, y1));

            ++x1;
            e -= dy;
            if (e <= 0)
            {
                ++y1;
                e += dx;
            }
        }

        image.save("bresenham.png");
        */
    }

    // version bicolore
    {
        QImage image(width, height, QImage::Format_ARGB32);

        QBENCHMARK
        {
            bresenhamHorizontal(image);
        }

        image.save("bresenham.png");
    }
}

void bresenhamVerticalInBuffer(uchar* pbuffer, int red, int green, int blue)
{
    int y = 0;
    const int y_end = height - 1;

    const int red_end = 255;
    const int green_end = 255;
    const int blue_end = 255;

    const int e_y = y_end - y;
    const int d_y = 2 * e_y;

    int e_red = e_y;
    int e_green = e_y;
    int e_blue = e_y;

    const int d_red = 2 * (red_end - red);
    const int d_green = 2 * (green_end - green);
    const int d_blue = 2 * (blue_end - blue);

    while (y <= y_end)
    {
        *pbuffer = 255 - red; // blue
        ++pbuffer;
        *pbuffer = 255 - green; // green
        ++pbuffer;
        *pbuffer = 255 - blue; // red
        ++pbuffer;
        *pbuffer = (uchar) 255; // alpha
        ++pbuffer;

        ++y;

        e_red -= d_red;
        if (e_red <= 0)
        {
            ++red;
            e_red += d_y;
        }

        e_green -= d_green;
        if (e_green <= 0)
        {
            ++green;
            e_green += d_y;
        }

        e_blue -= d_blue;
        if (e_blue <= 0)
        {
            ++blue;
            e_blue += d_y;
        }
    }
}

void bresenhamHorizontalInBuffer(uchar* pbuffer)
{
    int x = 0;
    const int x_end = width - 1;

    int red = 0;
    int green = 0;
    int blue = 0;

    const int red_end = 255 - main_color.red();
    const int green_end = 255 - main_color.green();
    const int blue_end = 255 - main_color.blue();

    const int e_x = x_end - x;
    const int d_x = 2 * e_x;

    int e_red = e_x;
    int e_green = e_x;
    int e_blue = e_x;

    const int d_red = 2 * (red_end - red);
    const int d_green = 2 * (green_end - green);
    const int d_blue = 2 * (blue_end - blue);

    while (x <= x_end)
    {
        bresenhamVerticalInBuffer(pbuffer, red, green, blue);

        ++x;

        e_red -= d_red;
        if (e_red <= 0)
        {
            ++red;
            e_red += d_x;
        }

        e_green -= d_green;
        if (e_green <= 0)
        {
            ++green;
            e_green += d_x;
        }

        e_blue -= d_blue;
        if (e_blue <= 0)
        {
            ++blue;
            e_blue += d_x;
        }
    }
}

void Benchmark::bresenhamInBuffer()
{
    uchar buffer[width * height * 4];
    QImage image(buffer, width, height, QImage::Format_ARGB32);

    QBENCHMARK
    {
        uchar* pbuffer = &buffer[0];
        bresenhamHorizontalInBuffer(pbuffer);
    }

    image.save("buffured.png");
}

QTEST_MAIN(Benchmark)
